# -*- coding: utf-8 -*-
# *****************************************************************************
#   Copyright 2017 Karl Einar Nelson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

# Optional jpype module to support:
#   import <java_pkg> [ as <name> ]
#   import <java_pkg>.<java_class> [ as <name> ]
#   from <java_pkg> import <java_class>[,<java_class>*]
#   from <java_pkg> import <java_class> [ as <name> ]
#   from <java_pkg>.<java_class> import <java_static> [ as <name> ]
#   from <java_pkg>.<java_class> import <java_inner> [ as <name> ]
#
#   jpype.imports.registerDomain(moduleName, alias=<java_pkg>)
#   jpype.imports.registerImportCustomizer(JImportCustomizer)
# 
# Requires Python 3.6 or later
# Usage:
#   import jpype
#   import jpype.imports
#   <start or attach jvm>
#   # Import java packages as modules
#   from java.lang import String

try:
    from importlib.machinery import ModuleSpec as _ModuleSpec
    from types import ModuleType as _ModuleType
except Exception:
    raise ImportError("jpype.imports Not supported for Python 2")
import sys as _sys
import keyword as _keyword
from ._jclass import JClass as _JClass
from ._jclass import _JavaClass as _JavaClass
from ._core import registerJVMInitializer as _jinit

__all__ = ["registerImportCustomizer", "registerDomain", "JImportCustomizer"]
_initialized = False
_exportTypes = ()
_modifier = None

# %% Utility
def _keywordUnwrap(name):
    if not name.endswith('_'):
        return name
    if _keyword.iskeyword(name[:-1]):
        return name[:-1]
    return name

def _keywordWrap(name):
    if name in _keyword.kwlist:
        return name + "_"
    return name

def _getJavaClass(javaname):
    try:
        return _JClass(javaname)
    except Exception:
        return None

def _copyProperties(out, mc):
    for v in dir(mc):
        # Skip private members
        if v.startswith('_'):
            continue

        # Copy properties
        attr = getattr(mc, v)
        if isinstance(attr, property):
            out[v] = attr

def _getStaticMethods(cls):
    global _modifier
    static = {}
    for u in cls.__javaclass__.getMethods():
        if not _modifier.isStatic(u.getModifiers()):
            continue
        name = _keywordWrap(u.getName())
        static[name] = getattr(cls, name)
    return static

def _copyStaticMethods(out, cls):
    for u, v in _getStaticMethods(cls).items():
        out[u] = v

# %% Customizer
_CUSTOMIZERS = []

def registerImportCustomizer(customizer):
    """ Import customizers can be used to import python packages
    into java modules automatically.
    """
    _CUSTOMIZERS.append(customizer)

# Support hook for placing other things into the java tree
class JImportCustomizer(object):
    """ Base class for Import customizer.

    Import customizers should implement canCustomize and getSpec.

    Example:
      | # Site packages for each java package are stored under $DEVEL/<java_pkg>/py
      | class SiteCustomizer(jpype.imports.JImportCustomizer):
      |     def canCustomize(self, name):
      |         if name.startswith('org.mysite') and name.endswith('.py'):
      |             return True
      |         return False
      |     def getSpec(self, name):
      |         pname = name[:-3]
      |         devel = os.environ.get('DEVEL')
      |         path = os.path.join(devel, pname,'py','__init__.py')
      |         return importlib.util.spec_from_file_location(name, path)
   """
    def canCustomize(self, name):
        """ Determine if this path is to be treated differently 

        Return:
            True if an alternative spec is required.
        """
        return False

    def getSpec(self, name):
        """ Get the module spec for this module. 
        """
        raise Exception

# %% Import
class _JImport(object):
    """ (internal) Base class for import java modules """
    # Module requirements
    __doc__ = None
    __loader__ = None
    __path__ = "<java>"
    __package__ = "java"

    def __init__(self, name):
        pass

    def __getattr__(self, name):
        if name.startswith('_'):
            return object.__getattribute__(self, name)

        name = _keywordUnwrap(name)

        # Inner class support
        jname = object.__getattribute__(self, '__javaname__')
        try:
            object.__getattribute__(self, '__javaclass__')
            jname = "$".join([jname, name])
        except AttributeError:
            jname = ".".join([jname, name])

        # Get the class (if it exists)
        jtype = _getJavaClass(jname)
        if jtype:
            # Cache it for later
            object.__setattr__(self, name, jtype)
            return jtype

        # If the java class does not exist, throw a ClassNotFound exception
        raise Exception("Unable to find java class " + jname)

    def __setattr__(self, name, value):
        if name.startswith('__'):
            raise AttributeError("Module does not allow setting of %s" % name)
        if hasattr(value, '__javaclass__'):
            return object.__setattr__(self, name, getattr(value, '__javaclass__'))
        if isinstance(value, (_JImport, _ModuleType)):
            return object.__setattr__(self, name, value)
        raise AttributeError("JImport may not set attribute %s" % name)


# In order to get properties to be attached to the _JImport class,
# we must create a dynamic class between
def _JImportFactory(spec, javaname, cls=_JImport):
    """ (internal) Factory for creating java modules dynamically.

    This is needed to create a new type node to hold static methods.
    """
    def init(self, name):
        # Call the base class
        cls.__init__(self, name)

    def getall(self):
        global _exportTypes
        d1 = self.__dict__.items()
        d2 = self.__class__.__dict__.items()
        local = [name for name, attr in d1 if not name.startswith('_')
                and isinstance(attr, _exportTypes)]
        glob = [name for name, attr in d2 if not name.startswith('_')
                and isinstance(attr, _exportTypes)]
        local.extend(glob)
        return local

    # Set up a new class for this type
    bases = [cls]
    members = {
        "__init__": init,
        "__javaname__": javaname,
        "__name__": spec.name,
        "__all__": property(getall),
        "__spec__": spec,
    }

    # Is this module also a class, if so insert class info
    jclass = _getJavaClass(javaname)
    if jclass:
        # Mark this as a class (will cause children to be inner classes)
        members['__javaclass__'] = jclass

        # Exposed static members as part of the module
        _copyProperties(members, jclass.__metaclass__)
        _copyStaticMethods(members, jclass)

    return type("module." + spec.name, tuple(bases), members)

def _JModule(spec, javaname):
    """ (internal) Front end for creating a java module dynamically """
    cls = _JImportFactory(spec, javaname)
    out = cls(spec.name)
    return out

# %% Finder
class _JImportLoader:
    """ (internal) Finder hook for importlib. """
    def find_spec(self, name, path, target):
        parts = name.split('.', 1)
        if not parts[0] in _JDOMAINS:
            return None

        # Support for external modules in java tree
        for customizer in _CUSTOMIZERS:
            if customizer.canCustomize(name):
                return customizer.getSpec(name)

        # Import the java module
        return _ModuleSpec(name, self)

    """ (internal) Loader hook for importlib. """
    def create_module(self, spec):
        global _initialized
        if not _initialized:
            raise ImportError("Attempt to create java modules without jvm")

        # Handle creating the java name based on the path
        parts = spec.name.split('.')
        if len(parts) == 1:
            return _JModule(spec, _JDOMAINS[spec.name])

        # Use the parent module to simplify name mangling
        base = _sys.modules[".".join(parts[:-1])]

        # Support of inner classes
        if not isinstance(base,_JImport):
            return getattr(base, parts[-1])
        jbasename = object.__getattribute__(base, '__javaname__')
        try:
            object.__getattribute(base, '__javaclass__')
            javaname = "$".join([jbasename, _keywordUnwrap(parts[-1])])
        except AttributeError:
            javaname = ".".join([jbasename, _keywordUnwrap(parts[-1])])

        return _JModule(spec, javaname)

    def exec_module(self, fullname):
        pass

# Install hooks into python importlib
_sys.meta_path.append(_JImportLoader())

# %% Domains
_JDOMAINS = {}

def registerDomain(mod, alias=None):
    """ Add a java domain to python as a dynamic module.

    Args:
        mod is the name of the dynamic module
        alias is the name of the java path. (optional)
    """
    if not alias:
        alias = mod
    _JDOMAINS[mod] = alias

# Preregister common top level domains
registerDomain('com')
registerDomain('gov')
registerDomain('java')
registerDomain('org')

# %% Initialize
def _initialize():
    global _exportTypes
    global _initialized 
    global _modifier
    _initialized = True
    _JMethod = type(_JClass('java.lang.Class').forName)
    _modifier = _JClass('java.lang.reflect.Modifier')
    _exportTypes = (property, _JavaClass, _JImport, _JMethod)

_jinit(_initialize)




#*****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************

import collections
import sys

import _jpype
from . import _jclass
from . import JClassUtil

if sys.version > '3':
    unicode = str

def _initialize():
    _jpype.setResource('ProxyClass',JProxy)

class JProxy(object):
    def __init__(self, intf, dict=None, inst=None):
        actualIntf = None

        if isinstance(intf, str) or isinstance(intf, unicode):
            actualIntf = [_jclass.JClass(intf)]
        elif isinstance(intf, _jclass._JavaClass):
            actualIntf = [intf]
        elif isinstance(intf, collections.Sequence):
            actualIntf = []
            for i in intf:
                if isinstance(i, str) or isinstance(i, unicode):
                    actualIntf.append(_jclass.JClass(i))
                elif isinstance(i, _jclass._JavaClass):
                    actualIntf.append(i)
                else:
                    raise TypeError("JProxy requires java interface classes "
                                    "or the names of java interfaces classes")
        else:
            raise TypeError("JProxy requires java interface classes "
                            "or the names of java interfaces classes")

        for i in actualIntf:
            if not JClassUtil.isInterface(i):
                raise TypeError("JProxy requires java interface classes "
                                "or the names of java interfaces classes: {0}"
                                .format(i.__name__))

        if dict is not None and inst is not None:
            raise RuntimeError("Specify only one of dict and inst")

        self._dict = dict
        self._inst = inst

        self._proxy = _jpype.createProxy(self, actualIntf)

    def getCallable(self, name):
        if self._dict is not None:
            return self._dict[name]
        else:
            return getattr(self._inst, name)



#*****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************
import sys

import _jpype

from . import _jclass
from . import _jarray
from . import _jwrapper
from . import _jproxy
from . import _jexception
from . import _jcollection
from . import _jobject
from . import _jio
from . import _properties
from . import nio
from . import reflect
from . import _refdaemon

_usePythonThreadForDaemon = False

def setUsePythonThreadForDeamon(v):
    global _usePythonThreadForDaemon
    _usePythonThreadForDaemon = v

_initializers=[]

def registerJVMInitializer(func):
    """Register a function to be called after jvm is started"""
    _initializers.append(func)

def _initialize():
    _properties._initialize()
    _jclass._initialize()
    _jarray._initialize()
    _jwrapper._initialize()
    _jproxy._initialize()
    _jexception._initialize()
    _jcollection._initialize()
    _jobject._initialize()
    nio._initialize()
    reflect._initialize()
    for func in _initializers:
        func()

def isJVMStarted() :
    return _jpype.isStarted()

def startJVM(jvm, *args):
    """
    Starts a Java Virtual Machine

    :param jvm:  Path to the jvm library file (libjvm.so, jvm.dll, ...)
    :param args: Arguments to give to the JVM
    """
    _jpype.startup(jvm, tuple(args), True)
    _initialize()

    # start the reference daemon thread
    if _usePythonThreadForDaemon:
        _refdaemon.startPython()
    else:
        _refdaemon.startJava()

def attachToJVM(jvm):
    _jpype.attach(jvm)
    _initialize()

def shutdownJVM():
    _refdaemon.stop()
    _jpype.shutdown()

def isThreadAttachedToJVM():
    return _jpype.isThreadAttachedToJVM()

def attachThreadToJVM():
    _jpype.attachThreadToJVM()

def detachThreadFromJVM():
    _jpype.detachThreadFromJVM()


def get_default_jvm_path():
    """
    Retrieves the path to the default or first found JVM library

    :return: The path to the JVM shared library file
    :raise ValueError: No JVM library found
    """
    if sys.platform == "cygwin":
        # Cygwin
        from ._cygwin import WindowsJVMFinder
        finder = WindowsJVMFinder()
    elif sys.platform == "win32":
        # Windows
        from ._windows import WindowsJVMFinder
        finder = WindowsJVMFinder()
    elif sys.platform == "darwin":
        # Mac OS X
        from ._darwin import DarwinJVMFinder
        finder = DarwinJVMFinder()
    else:
        # Use the Linux way for other systems
        from ._linux import LinuxJVMFinder
        finder = LinuxJVMFinder()

    return finder.get_jvm_path()

# Naming compatibility
getDefaultJVMPath = get_default_jvm_path


class ConversionConfigClass(object):
    def __init__(self):
        self._convertString = 1

    def _getConvertString(self):
        return self._convertString

    def _setConvertString(self, value):
        if value:
            self._convertString = 1
        else:
            self._convertString = 0

        _jpype.setConvertStringObjects(self._convertString)

    string = property(_getConvertString, _setConvertString, None)

ConversionConfig = ConversionConfigClass()



#*****************************************************************************
#   Copyright 2017 Karl Einar Nelson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************
from . import _jclass
from . import JavaException
import sys as _sys

if _sys.version > '3':
    pass

def _closeableExit(self,exception_type, exception_value, traceback):
    info = _sys.exc_info()
    try:
        self.close()
    except JavaException as jex:
        # Eat the second exception if we are already handling one.
        if (info[0]==None):
            raise jex
        pass

def _closeableEnter(self):
    return self


class CloseableCustomizer(object):
    _METHODS = {
        '__enter__': _closeableEnter,
        '__exit__': _closeableExit,
    }

    def canCustomize(self, name, jc):
        if name == 'java.io.Closeable':
            return True
        return False

    def customize(self, name, jc, bases, members):
        members.update(CloseableCustomizer._METHODS)

_jclass.registerClassCustomizer(CloseableCustomizer())



#*****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************

from . import _jvmfinder
from ._jvmfinder import JVMNotSupportedException
import struct as _struct
import sys as _sys

# ------------------------------------------------------------------------------

def _checkJVMArch(jvmPath):
    IMAGE_FILE_MACHINE_I386=332
    IMAGE_FILE_MACHINE_IA64=512
    IMAGE_FILE_MACHINE_AMD64=34404

    is64 = _sys.maxsize > 2**32
    with open(jvmPath, "rb") as f:
        s=f.read(2)
        if s!=b"MZ":
            raise JVMNotSupportedException("JVM not valid")
        else:
            f.seek(60)
            s=f.read(4)
            header_offset=_struct.unpack("<L", s)[0]
            f.seek(header_offset+4)
            s=f.read(2)
            machine=_struct.unpack("<H", s)[0]

    if machine==IMAGE_FILE_MACHINE_I386:
        if is64:
            raise JVMNotSupportedException("JVM mismatch, python is 64 bit and JVM is 32 bit.")
    elif machine==IMAGE_FILE_MACHINE_IA64 or machine==IMAGE_FILE_MACHINE_AMD64:
        if not is64:
            raise JVMNotSupportedException("JVM mismatch, python is 32 bit and JVM is 64 bit.")
    else:
        raise JVMNotSupportedException("Unable to deterime JVM Type")

class WindowsJVMFinder(_jvmfinder.JVMFinder):
    """
    Windows JVM library finder class
    """
    def __init__(self):
        """
        Sets up members
        """
        # Call the parent constructor
        _jvmfinder.JVMFinder.__init__(self)

        # Library file name
        self._libfile = "jvm.dll"

        # Search methods
        winreg = self._get_winreg()
        if winreg!=None:
            self._methods = (self._get_from_java_home, self._get_from_registry)
        else:
            self._methods = (self._get_from_java_home, )

    def check(self, jvm):
        _checkJVMArch(jvm)

    def _get_winreg(self):
        # Py2
        try:
            import _winreg as winreg
            return winreg
        except ImportError:
            pass

        # in Py3, winreg has been moved
        try:
            import winreg
            return winreg
        except ImportError:
            pass
        return None


    def _get_from_registry(self):
        """
        Retrieves the path to the default Java installation stored in the
        Windows registry

        :return: The path found in the registry, or None
        """
        try:
            # Winreg is an optional package in cygwin
            winreg = self._get_winreg()
            if not winreg:
                return None

            jreKey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                r"SOFTWARE\JavaSoft\Java Runtime Environment")
            cv = winreg.QueryValueEx(jreKey, "CurrentVersion")
            versionKey = winreg.OpenKey(jreKey, cv[0])
            winreg.CloseKey(jreKey)

            cv = winreg.QueryValueEx(versionKey, "RuntimeLib")
            winreg.CloseKey(versionKey)

            return cv[0]

        except WindowsError:
            return None




import jpype
from ._jclass import _JavaObject
from ._jclass import _isJavaCtor
from ._jwrapper import _JWrapper
import sys as _sys

#Python2/3 support
if _sys.version_info > (3,):
    _long  =  int
else:
    _long  =  long

__all__=[]
   
# Temporary JavaObject wrapper to call Java Methods in __new__
class _JTmp(_JavaObject):
    def __init__(self, jp):
        self.__javaobject__ = jp

class _JBoxedShort(int):
    __metaclass__ = int.__class__
    def __new__(cls, *args, **kwargs):
        try:
            if isinstance(args[0], (int, _long)):
                return int.__new__(cls, args[0])
            if isinstance(args[0], _JWrapper):
                return int.__new__(cls, args[0]._pyv)
        except OverflowError:
            raise TypeError("Overflow")
        if _isJavaCtor(args):
            return int.__new__(cls, cls.shortValue(_JTmp(args[0][1])))
        raise ValueError("Invalid arguments %s"%args[0])

class _JBoxedInteger(int):
    __metaclass__ = int.__class__
    def __new__(cls, *args, **kwargs):
        try:
            if isinstance(args[0], (int, _long)):
                return int.__new__(cls, args[0])
            if isinstance(args[0], _JWrapper):
                return int.__new__(cls, args[0]._pyv)
        except OverflowError:
            raise TypeError("Overflow")
        if _isJavaCtor(args):
            return int.__new__(cls, cls.intValue(_JTmp(args[0][1])))
        raise ValueError("Invalid arguments")

class _JBoxedLong(_long):
    __metaclass__ = _long.__class__
    def __new__(cls, *args, **kwargs):
        if isinstance(args[0], (int, _long)):
            return _long.__new__(cls, args[0])
        if isinstance(args[0], _JWrapper):
            return _long.__new__(cls, args[0]._pyv)
        if _isJavaCtor(args):
            return _long.__new__(cls, cls.longValue(_JTmp(args[0][1])))
        raise ValueError("Invalid arguments")

class _JBoxedFloat(float):
    __metaclass__ = float.__class__
    def __new__(cls, *args, **kwargs):
        if isinstance(args[0], (int, _long,float)):
            return float.__new__(cls, args[0])
        if isinstance(args[0], _JWrapper):
            return int.__new__(cls, args[0]._pyv)
        if _isJavaCtor(args):
            return float.__new__(cls, cls.floatValue(_JTmp(args[0][1])))
        raise ValueError("Invalid arguments")

class _JBoxedDouble(float):
    __metaclass__ = float.__class__
    def __new__(cls, *args, **kwargs):
        if isinstance(args[0], (int, _long,float)):
            return float.__new__(cls, args[0])
        if isinstance(args[0], _JWrapper):
            return int.__new__(cls, args[0]._pyv)
        if _isJavaCtor(args):
            return float.__new__(cls, cls.doubleValue(_JTmp(args[0][1])))
        raise ValueError("Invalid arguments")

_BOXED = {
        'java.lang.Short': _JBoxedShort,
        'java.lang.Integer': _JBoxedInteger,
        'java.lang.Long': _JBoxedLong,
        'java.lang.Float': _JBoxedFloat,
        'java.lang.Double': _JBoxedDouble,
        }

class _BoxedCustomizer(object):
    def canCustomize(self, name, jc):
        if name in _BOXED:
            return True
        return False

    def customize(self, name, jc, bases, members):
        cls=_BOXED[name]
        bases.append(cls)
        members['__eq__']=cls.__eq__
        members['__ne__']=cls.__ne__
        if hasattr(cls, '__cmp__'):
            members['__cmp__']=cls.__cmp__

def _initialize():
    from ._jclass import registerClassCustomizer
    registerClassCustomizer(_BoxedCustomizer())

_initialize()



#*****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************

import _jpype
from . import _jclass

class JPackage(object):
    def __init__(self, name):
        self.__name = name

    def __getattribute__(self, n):
        try:
            return object.__getattribute__(self, n)
        except AttributeError as ex:
            if n.startswith("__"):
                raise ex
            # not found ...

            # perhaps it is a class?
            subname = "{0}.{1}".format(self.__name, n)
            if not _jpype.isStarted():
               import warnings
               warnings.warn("JVM not started yet, can not inspect JPackage contents")
               return n
            if not _jpype.isThreadAttachedToJVM():
                _jpype.attachThreadToJVM()
            cc = _jpype.findClass(subname)
            if cc is None:
                # can only assume it is a sub-package then ...
                cc = JPackage(subname)
            else:
                cc = _jclass._getClassFor(cc)

            self.__setattr__(n, cc, True)
            return cc

    def __setattr__(self, n, v, intern=False):
        if not n[:len('_JPackage')] == '_JPackage' \
           and not intern: # NOTE this shadows name mangling
            raise RuntimeError("Cannot set attributes in a package {0}"
                               .format(n))
        object.__setattr__(self, n, v)

    def __str__(self):
        return "<Java package {0}>".format(self.__name)

    def __call__(self, *arg, **kwarg):
        raise TypeError("Package {0} is not Callable".format(self.__name))



#*****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************
import collections
import sys

import _jpype
from . import _jclass
from . import _jwrapper

if sys.version > '3':
    unicode = str
    irange = range

_CLASSES = {}
_CUSTOMIZERS = []

def _initialize():
    _jpype.setResource('JavaArrayClass',_JavaArrayClass)
    _jpype.setResource('GetJavaArrayClassMethod',_getClassFor)

    registerArrayCustomizer(CharArrayCustomizer())
    registerArrayCustomizer(ByteArrayCustomizer())

def registerArrayCustomizer(c):
    _CUSTOMIZERS.append(c)

class _JavaArrayClass(object):
    def __init__(self, jo):
        self.__javaobject__ = jo

    def __str__(self):
        return str(tuple(self))

    def __len__(self):
        return _jpype.getArrayLength(self.__javaobject__)

    def __iter__(self):
        return _JavaArrayIter(self)

    def __getitem__(self, ndx):
        if isinstance(ndx, slice):
            start, stop, step = ndx.indices(len(self))
            if step != 1:
                raise RuntimeError("Slicing with step unimplemented")
            return self.__getslice__(start, stop)
        return _jpype.getArrayItem(self.__javaobject__, ndx)

    def __setitem__(self, ndx, val):
        if isinstance(ndx, slice):
            start, stop, step = ndx.indices(len(self))
            if step != 1:
                # Iterate in python if we need to step
                indices = irange(start, stop, step)
                for index, value in zip(indices, val):
                    self[index] = value
            else:
                self.__setslice__(start, stop, val)
            return
        _jpype.setArrayItem(self.__javaobject__, ndx, val)

    def __getslice__(self, i, j):
        if j == sys.maxsize:
            j = _jpype.getArrayLength(self.__javaobject__)
        return _jpype.getArraySlice(self.__javaobject__, i, j)

    def __setslice__(self, i, j, v):
        if j == sys.maxsize:
            j = _jpype.getArrayLength(self.__javaobject__)
        _jpype.setArraySlice(self.__javaobject__, i, j, v)

    def __setattr__(self, attr, value):
        if attr.startswith('_') \
               or callable(value) \
               or isinstance(getattr(self.__class__, attr), property):
            object.__setattr__(self, attr, value)
        else:
            raise AttributeError("%s does not have field %s"%(self.__name__, attr), self)


def _jarrayInit(self, *args):
    if len(args) == 2 and args[0] == _jclass._SPECIAL_CONSTRUCTOR_KEY:
        _JavaArrayClass.__init__(self, args[1])
    elif len(args) != 1:
        raise TypeError(
            "Array classes only take 2 parameters, {0} given"
            .format(len(args) + 1))
    else:
        values = None
        if isinstance(args[0], collections.Sequence):
            sz = len(args[0])
            values = args[0]
        else:
            sz = args[0]

        _JavaArrayClass.__init__(
            self, _jpype.newArray(self.__class__.__javaclass__, sz))

        if values is not None:
            _jpype.setArraySlice(self.__javaobject__, 0, sz, values)


class _JavaArrayIter(object):
    def __init__(self, a):
        self._array = a
        self._ndx = -1

    def __iter__(self):
        return self

    def __next__(self):
        self._ndx += 1
        if self._ndx >= len(self._array):
            raise StopIteration
        return self._array[self._ndx]

    next = __next__

class _JavaArray(type):
    pass

def _defineArrayClass(name, jt):
    members = {
            "__init__": _jarrayInit,
            "__javaclass__": jt,
            "__name__": name,
    }

    bases = [_JavaArrayClass]

    for i in _CUSTOMIZERS:
        if i.canCustomize(name, jt):
            i.customize(name, jt, bases, members)

    return _JavaArray(name, tuple(bases), members)

def _getClassFor(name):
    if name not in _CLASSES:
        jc = _jpype.findArrayClass(name)
        _CLASSES[name] = _defineArrayClass(name, jc)

    return _CLASSES[name]


def JArray(t, ndims=1):
    if issubclass(t, _jwrapper._JWrapper):
        t = t.typeName

    elif isinstance(t, _JavaArray):
        t = t.typeName

    elif issubclass(t, _jclass._JAVAOBJECT):
        t = t.__name__

    elif not isinstance(t, str) and not isinstance(t, unicode):
        raise TypeError("Argument must be a java class, java array class, "
                        "java wrapper or string representing a java class")

    arrayTypeName = t + ('[]' * ndims)

    return _getClassFor(arrayTypeName)

def _charArrayStr(self):
    return ''.join(self)

def _charArrayUnicode(self):
    return u''.join(self)

class CharArrayCustomizer(object):
    def canCustomize(self, name, jc):
        if name == 'char[]':
            return True
        return False

    def customize(self, name, jc, bases, members):
        members['__str__'] = _charArrayStr
        members['__unicode__'] = _charArrayUnicode

def _byteArrayStr(self):
    s = _jclass.JClass('java.lang.String')(self)
    return s.toString()


class ByteArrayCustomizer(object):
    def canCustomize(self, name, jc):
        if name == 'byte[]':
            return True
        return False

    def customize(self, name, jc, bases, members):
        members['__str__'] = _byteArrayStr



#*****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#*****************************************************************************
import _jpype
from ._pykeywords import KEYWORDS



_CLASSES = {}

_SPECIAL_CONSTRUCTOR_KEY = "This is the special constructor key"

_JAVACLASS = None
_JAVAOBJECT = None
_JAVATHROWABLE = None
_COMPARABLE = None
_RUNTIMEEXCEPTION = None

_CUSTOMIZERS = []

_COMPARABLE_METHODS = {
        "__cmp__": lambda self, o: self.compareTo(o)
        }

def _initialize():
    global _COMPARABLE, _JAVACLASS, _JAVAOBJECT, _JAVATHROWABLE, _RUNTIMEEXCEPTION
    registerClassCustomizer(_JavaLangClassCustomizer())

    _JAVAOBJECT = JClass("java.lang.Object")
    _JAVACLASS = JClass("java.lang.Class")
    _JAVATHROWABLE = JClass("java.lang.Throwable")
    _RUNTIMEEXCEPTION = JClass("java.lang.RuntimeException")
    _jpype.setResource('JavaClass', _JavaClass)
    _jpype.setResource('JavaObject', _JavaObject)
    _jpype.setResource('GetClassMethod',_getClassFor)
    _jpype.setResource('SpecialConstructorKey',_SPECIAL_CONSTRUCTOR_KEY)


def registerClassCustomizer(c):
    _CUSTOMIZERS.append(c)

class JClassCustomizer(object):
    def canCustomize(self, name):
        """ Deterimine if this class can be customized by this customizer.

        Return true if customize should be called, false otherwise."""
        pass

    def customize(self, name, jc, bases, members=None, fields=None, **kwargs):
        """ Customize the class.  

        Should be able to handle keyword arguments to support changes in customizers.
        """
        pass
    

def JClass(name):
    jc = _jpype.findClass(name)
    if jc is None:
        raise _RUNTIMEEXCEPTION.PYEXC("Class %s not found" % name)

    return _getClassFor(jc)


def _getClassFor(javaClass):
    name = javaClass.getName()
    if name in _CLASSES:
        return _CLASSES[name]

    pyJavaClass = _JavaClass(javaClass)
    _CLASSES[name] = pyJavaClass
    return pyJavaClass


def _javaNew(self, *args):
    return object.__new__(self)


def _javaExceptionNew(self, *args):
    return Exception.__new__(self)


def _isJavaCtor(args):
    if len(args)==1 and isinstance(args[0], tuple) \
            and args[0][0] is _SPECIAL_CONSTRUCTOR_KEY:
        return True
    return False
 

def _javaInit(self, *args):
    object.__init__(self)

    if len(args) == 1 and isinstance(args[0], tuple) \
       and args[0][0] is _SPECIAL_CONSTRUCTOR_KEY:
        self.__javaobject__ = args[0][1]
    else:
        self.__javaobject__ = self.__class__.__javaclass__.newClassInstance(
            *args)


def _javaGetAttr(self, name):
    try:
        r = object.__getattribute__(self, name)
    except AttributeError as ex:
        if name in dir(self.__class__.__metaclass__):
            r = object.__getattribute__(self.__class__, name)
        else:
            raise ex

    if isinstance(r, _jpype._JavaMethod):
        return _jpype._JavaBoundMethod(r, self)
    return r

def _javaSetAttr(self, attr, value):
    if attr.startswith('_') \
           or callable(value) \
           or isinstance(getattr(self.__class__, attr), property):
        object.__setattr__(self, attr, value)
    else:
        raise AttributeError("%s does not have field %s"%(self.__name__, attr), self)

def _mro_override_topsort(cls):
    # here we run a topological sort to get a linear ordering of the inheritance graph.
    parents = set().union(*[x.__mro__ for x in cls.__bases__])
    numsubs = dict()
    for cls1 in parents:
        numsubs[cls1] = len([cls2 for cls2 in parents if cls1 != cls2 and issubclass(cls2,cls1)])
    mergedmro = [cls]
    while numsubs:
        for k1,v1 in numsubs.items():
            if v1 != 0: continue
            mergedmro.append(k1)
            for k2,v2 in numsubs.items():
                if issubclass(k1,k2):
                    numsubs[k2] = v2-1
            del numsubs[k1]
            break
    return mergedmro

class _MetaClassForMroOverride(type):
    def mro(cls):
        return _mro_override_topsort(cls)


class _JavaObject(object):
    """ Base class for all Java Objects. 

        Use isinstance(obj, jpype.JavaObject) to test for a object.
    """
    pass

#  JPype has several class types (assuming Foo is a java class)
#     Foo$$Static - python meta class for Foo holding
#       properties for static fields and static methods
#
#     Foo - Python class which produces a Foo object() 
#       and access to static fields and stataic methods
#       in addition as a class type it holds all the fields and methods 
#       inherites from _JavaClass
#     Foo.__javaclass__ - private jpype capsule holding C resources
#     Foo.__class__ - python class type for class (will be Foo$$Static)
#     Foo.class_ - java.lang.Class<Foo>
#
#     Foo() - instance of Foo which wraps a java object
#       inherits from _JavaObject
#     Foo().__class__ - ptthon class type for object (will be Foo)
#     Foo().getClass() - java.lang.Class<Foo> 
#


class _JavaClass(type):
    """ Base class for all Java Class types. 

        Use isinstance(obj, jpype.JavaClass) to test for a class.
    """
    def __new__(cls, jc):
        global _JAVACLASS
        bases = []
        name = jc.getName()

        static_fields = {}
        members = {
                "__javaclass__": jc,
                "__init__": _javaInit,
                "__str__": lambda self: self.toString(),
                "__hash__": lambda self: self.hashCode(),
                "__eq__": lambda self, o: self.equals(o),
                "__ne__": lambda self, o: not self.equals(o),
                "__getattribute__": _javaGetAttr,
                "__setattr__": _javaSetAttr,
                }

        if name == 'java.lang.Object':
            bases.append(_JavaObject)
        elif jc.isPrimitive():
            bases.append(object)
        elif not jc.isInterface():
            bjc = jc.getBaseClass()
            bases.append(_getClassFor(bjc))

        if _JAVATHROWABLE is not None and jc.isSubclass("java.lang.Throwable"):
            from . import _jexception
            members["PYEXC"] = _jexception._makePythonException(name, bjc)

        itf = jc.getBaseInterfaces()
        for ic in itf:
            bases.append(_getClassFor(ic))

        if len(bases) == 0:
            bases.append(_JavaObject)

        # add the fields
        fields = jc.getClassFields()
        for i in fields:
            fname = i.getName()
            if fname in KEYWORDS:
                fname += "_"

            if i.isStatic():
                g = lambda self, fld=i: fld.getStaticAttribute()
                s = None
                if not i.isFinal():
                    s = lambda self, v, fld=i: fld.setStaticAttribute(v)
                static_fields[fname] = property(g, s)
            else:
                g = lambda self, fld=i: fld.getInstanceAttribute(
                    self.__javaobject__)
                s = None
                if not i.isFinal():
                    s = lambda self, v, fld=i: fld.setInstanceAttribute(
                        self.__javaobject__, v)
                members[fname] = property(g, s)

        # methods
        methods = jc.getClassMethods()  # Return tuple of tuple (name, method).
        for jm in methods:
            mname = jm.getName()
            if mname in KEYWORDS:
                mname += "_"

            members[mname] = jm

        static_fields['mro'] = _mro_override_topsort
        static_fields['class_']= property(lambda self: _JAVACLASS.forName(name,True, JClass('java.lang.ClassLoader').getSystemClassLoader()), None)

        for i in _CUSTOMIZERS:
            if i.canCustomize(name, jc):
                if isinstance(i, JClassCustomizer):
                    i.customize(name, jc, bases, members=members, fields=static_fields)
                else:
                    i.customize(name, jc, bases, members)

        # Prepare the meta-metaclass
        meta_bases = []
        for i in bases:
            if i is object or i is _JavaObject:
                meta_bases.append(cls)
            else:
                meta_bases.append(i.__metaclass__)

        metaclass = type.__new__(_MetaClassForMroOverride, name + "$$Static", tuple(meta_bases),
                                 static_fields)
        members['__metaclass__'] = metaclass
        result = type.__new__(metaclass, name, tuple(bases), members)

        return result

# Patch for forName
def _jclass_forName(self, *args):
    if len(args)==1 and isinstance(args[0],str):
        return self._forName(args[0], True, JClass('java.lang.ClassLoader').getSystemClassLoader())
    else:
        return self._forName(*args)

class _JavaLangClassCustomizer(JClassCustomizer):
    def canCustomize(self, name, jc):
        return name == 'java.lang.Class'

    def customize(self, name, jc, bases, members, fields, **kwargs):
        members['_forName']=members['forName']
        del members['forName']
        fields['forName']= _jclass_forName



import os as _os
import sys as _sys
import glob as _glob

__all__=['addClassPath', 'getClassPath']

_CLASSPATHS=set()
_SEP = _os.path.pathsep
if _sys.platform=='cygwin':
    _SEP=';'

def _init():
    global _CLASSPATHS
    global _SEP
    classpath=_os.environ.get("CLASSPATH")
    if classpath:
        _CLASSPATHS|=set(classpath.split(_SEP))

_init()

# Cygwin needs to convert to windows paths
if _sys.platform=='cygwin':
    _root=None
    def _get_root():
        global _root
        if _root!=None:
            return _root
        import subprocess
        proc = subprocess.Popen("cygpath -wa /", shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT, close_fds=True)
        _root=proc.stdout.read().strip().decode('utf-8')
        return _root

    def _splitpath(path):
        parts=[]
        (path, tail)=_os.path.split( path)
        while path and tail:
             parts.insert(0,tail)
             (path,tail)=_os.path.split(path)
        return parts

    def _posix2win(directory):
        root=_get_root()
        directory=_os.path.abspath(directory)
        paths=_splitpath(directory)
        if paths[0]=="cygdrive":
            paths.pop(0)
            drive=paths.pop(0)
            paths.insert(0, "%s:"%drive)
            return '\\'.join(paths)
        paths.insert(0,root)
        return '\\'.join(paths)
    # needed for testing
    __all__.append("_posix2win")

def addClassPath(path1):
    """ Add a path to the java class path"""
    global _CLASSPATHS
    path1=_os.path.abspath(path1)
    if _sys.platform=='cygwin':
        path1=_posix2win(path1)
    _CLASSPATHS.add(str(path1))

def getClassPath():
    """ Get the full java class path.

    Includes user added paths and the environment CLASSPATH.
    """
    global _CLASSPATHS
    global _SEP
    out=[]
    for path in _CLASSPATHS:
        if path=='':
            continue
        if path.endswith('*'):
            paths=_glob.glob(path+".jar")
            if len(path)==0:
                continue 
            out.extend(paths)
        else:
            out.append(path)
    return _SEP.join(out)


#print(getClassPath())



class WindowAdapter(object) :
    def __init__(self, *mth, **kw) :
        object.__init__(self)

        for i, j in kw.items() :
            setattr(self, i, j)

    def windowActivated(self, e) :
        pass

    def windowClosed(self,  e) :
        pass

    def windowClosing(self,  e) :
        pass

    def windowDeactivated(self,  e) :
        pass

    def windowDeiconified(self,  e) :
        pass

    def windowIconified(self,  e) :
        pass

    def windowOpened(self,  e) :
        pass



import sys
from jpype import javax, JObject

__JMenuBar = javax.swing.JMenuBar
__JMenu = javax.swing.JMenu

def buildMenuBar(menuDef):
    mb = __JMenuBar()

    for i in menuDef:
        jm = buildMenu(i[0], i[1])
        mb.add(JObject(jm, __JMenu))

    return mb

def buildMenu(name, menuDef):
    jm = __JMenu(name)

    for i in menuDef:
        if i is None:
            jm.addSeparator()
        elif isinstance(i, list) or isinstance(i, tuple):
            jm2 = buildMenu(i[0], i[1])
            jm.add(jm2)
        else:
            jm.add(i.proxy)

    return jm
